;;;; AJL - gps (strips), v2
;;;  MM, 2009

;;; Zalozene na:
;;;   Code from Paradigms of Artificial Intelligence Programming
;;;   http://norvig.com/paip/
;;;   Copyright (c) 1991 Peter Norvig
;;;
;;;   File gps1.lisp: First version of GPS (General Problem Solver)
;;;   File gps.lisp: Final version of GPS
;;;   File overview.lisp: miscellaneous functions from Overview chapter

;;;; Zadanie: Dostat syna do skoly
;;;     
;;;  K dispozicii su akcie, ktore je mozne vykonat. Kazda ma dane
;;;  podmienky, za ktorych moze byt vykonana, a jej vykonanie
;;;  vedie k zmene aktualneho stavu sveta.

;;;; Verzia 1 ******************************************************

(defstruct op "An operation"
	   (action nil) (preconds nil) (add-list nil) (del-list nil))

(defvar *school-ops*)

(setf *school-ops*
      (list (make-op :action 'drive-son-to-school
		     :preconds '(son-at-home car-works)
		     :add-list '(son-at-school)
		     :del-list '(son-at-home))
	    (make-op :action 'walk-son-to-school
		     :preconds '(son-at-home)
		     :add-list '(son-at-school)
		     :del-list '(son-at-home))
	    (make-op :action 'shop-installs-battery
		     :preconds '(car-needs-battery shop-knows-problem
				 shop-has-money)
		     :add-list '(car-works))
	    (make-op :action 'tell-shop-problem
		     :preconds '(in-communication-with-shop)
		     :add-list '(shop-knows-problem))
	    (make-op :action 'telephone-shop
		     :preconds '(know-phone-number)
		     :add-list '(in-communication-with-shop))
	    (make-op :action 'look-up-number
		     :preconds '(have-phone-book)
		     :add-list '(know-phone-number))
	    (make-op :action 'give-shop-money
		     :preconds '(have-money)
		     :add-list '(shop-has-money)
		     :del-list '(have-money))
	    (make-op :action 'ask-phone-number
		     :preconds '(in-communication-with-shop)
		     :add-list '(know-phone-number))))


(defun find-all (item sequence &rest keyword-args
                 &key (test #'eql) test-not &allow-other-keys)
  "Find all those elements of sequence that match item,
  according to the keywords.  Doesn't alter sequence."
  (if test-not
      (apply #'remove item sequence 
             :test-not (complement test-not) keyword-args)
      (apply #'remove item sequence
             :test (complement test) keyword-args)))

(defvar *state* nil "The current state: a list of conditions.")

(defvar *ops* nil "A list of available operators.")

(defun GPS (*state* goals *ops*)
  "General Problem Solver: achieve all goals using *ops*."
  (if (every #'achieve goals) 'solved))

(defun achieve (goal)
  "A goal is achieved if it already holds,
  or if there is an appropriate op for it that is applicable."
  (or (member goal *state*)
      (some #'apply-op 
            (find-all goal *ops* :test #'appropriate-p))))

(defun appropriate-p (goal op)
  "An op is appropriate to a goal if it is in its add list."
  (member goal (op-add-list op)))

(defun apply-op (op)
  "Print a message and update *state* if op is applicable."
  (when (every #'achieve (op-preconds op))
    (print (list 'executing (op-action op)))
    (setf *state* (set-difference *state* (op-del-list op)))
    (setf *state* (union *state* (op-add-list op)))
    t))

;;;; Verzia 2 ******************************************************

(defun executing-p (x)
  "Is x of the form: (executing ...) ?"
  (starts-with x 'executing))

(defun starts-with (list x)
  "Is this a list whose first element is x?"
  (and (consp list) (eql (first list) x)))

(defun convert-op (op)
  "Make op conform to the (EXECUTING op) convention."
  (unless (some #'executing-p (op-add-list op))
    (push (list 'executing (op-action op)) (op-add-list op)))
  op)

(defun op (action &key preconds add-list del-list)
  "Make a new operator that obeys the (EXECUTING op) convention."
  (convert-op
   (make-op :action action :preconds preconds
	    :add-list add-list :del-list del-list)))

(setf *school-ops*
      (list (op 'drive-son-to-school
		     :preconds '(son-at-home car-works)
		     :add-list '(son-at-school)
		     :del-list '(son-at-home))
	    (op 'walk-son-to-school
		     :preconds '(son-at-home)
		     :add-list '(son-at-school)
		     :del-list '(son-at-home))
	    (op 'shop-installs-battery
		     :preconds '(car-needs-battery shop-knows-problem
				 shop-has-money)
		     :add-list '(car-works))
	    (op 'tell-shop-problem
		     :preconds '(in-communication-with-shop)
		     :add-list '(shop-knows-problem))
	    (op 'telephone-shop
		     :preconds '(know-phone-number)
		     :add-list '(in-communication-with-shop))
	    (op 'look-up-number
		     :preconds '(have-phone-book)
		     :add-list '(know-phone-number))
	    (op 'give-shop-money
		     :preconds '(have-money)
		     :add-list '(shop-has-money)
		     :del-list '(have-money))
	    (op 'ask-phone-number
		     :preconds '(in-communication-with-shop)
		     :add-list '(know-phone-number))))

(defun achieve (goal)
  "A goal is achieved if it already holds,
  or if there is an appropriate op for it that is applicable."
  (or (member goal *state* :test #'equal)
      (some #'apply-op 
            (find-all goal *ops* :test #'appropriate-p))))

(defun appropriate-p (goal op)
  "An op is appropriate to a goal if it is in its add list."
  (member goal (op-add-list op) :test #'equal))

;;;; Verzia 3 ******************************************************

(defun GPS (state goals &optional (*ops* *ops*))
  "General Problem Solver: from state, achieve goals using *ops*."
  (remove-if #'atom
	     (let ((current-state (cons '(start) state)))
	       (if (every #'(lambda (g)
			      (setf current-state
				    (achieve current-state g)))
			  goals)
		   current-state))))

(defun achieve (state goal)
  "A goal is achieved if it already holds,
  or if there is an appropriate op for it that is applicable."
  (cond ((member goal state :test #'equal) state)
        (t (some #'(lambda (op) (apply-op state op))
                 (find-all goal *ops* :test #'appropriate-p)))))

(defun apply-op (state op)
  "Return a new, transformed state if op is applicable."
  (let ((state2 state))
    (mapc #'(lambda (g) (setf state2 (achieve state2 g)))
	  (op-preconds op))
    (unless (null state2)
      ;; Return an updated state
      (append (remove-if #'(lambda (x) 
                             (member x (op-del-list op) :test #'equal))
                         state2)
              (op-add-list op)))))

;;;; Verzia 4 ******************************************************

(defun GPS (state goals &optional (*ops* *ops*))
  "General Problem Solver: from state, achieve goals using *ops*."
  (remove-if #'atom (achieve-all (cons '(start) state) goals)))

(defun achieve-all (state goals)
  "Achieve each goal, and make sure they still hold at the end."
  (let ((current-state state))
    (if (and (every #'(lambda (g)
                        (setf current-state
                              (achieve current-state g)))
                    goals)
             (subsetp goals current-state :test #'equal))
        current-state)))

(defun apply-op (state op)
  "Return a new, transformed state if op is applicable."
  (let ((state2 (achieve-all state (op-preconds op))))
    (unless (null state2)
      ;; Return an updated state
      (append (remove-if #'(lambda (x) 
                             (member x (op-del-list op) :test #'equal))
                         state2)
              (op-add-list op)))))

;;;; Verzia 5 ******************************************************

(defun GPS (state goals &optional (*ops* *ops*))
  "General Problem Solver: from state, achieve goals using *ops*."
  (remove-if #'atom (achieve-all (cons '(start) state) goals nil)))

(defun achieve-all (state goals goal-stack)
  "Achieve each goal, and make sure they still hold at the end."
  (let ((current-state state))
    (if (and (every #'(lambda (g)
                        (setf current-state
                              (achieve current-state g goal-stack)))
                    goals)
             (subsetp goals current-state :test #'equal))
        current-state)))

(defun achieve (state goal goal-stack)
  "A goal is achieved if it already holds,
  or if there is an appropriate op for it that is applicable."
  (cond ((member goal state :test #'equal) state)
        ((member goal goal-stack :test #'equal) nil)
        (t (some #'(lambda (op) (apply-op state goal op goal-stack))
                 (find-all goal *ops* :test #'appropriate-p)))))

(defun apply-op (state goal op goal-stack)
  "Return a new, transformed state if op is applicable."
  (let ((state2 (achieve-all state (op-preconds op) 
                             (cons goal goal-stack))))
    (unless (null state2)
      ;; Return an updated state
      (append (remove-if #'(lambda (x) 
                             (member x (op-del-list op) :test #'equal))
                         state2)
              (op-add-list op)))))

;;;; Verzia 6 ******************************************************

(defvar *dbg-ids* nil "Identifiers used by dbg")

(defun ondebug (&rest ids)
  "Start dbg output on the given ids"
  (setf *dbg-ids* (union ids *dbg-ids*)))

(defun undebug (&rest ids)
  "Stpo dbg on the ids. With no ids, stop dbg altogether"
  (setf *dbg-ids* (if (null ids) nil
		      (set-difference *dbg-ids* ids))))

(defun dbg-indent (id indent format-string &rest args)
  "Print indented debugging info if (DEBUG ID) has been specified."
  (when (member id *dbg-ids*)
    (fresh-line *debug-io*)
    (dotimes (i indent) (princ "  " *debug-io*))
    (apply #'format *debug-io* format-string args)))


(defun achieve (state goal goal-stack)
  "A goal is achieved if it already holds,
  or if there is an appropriate op for it that is applicable."
  (dbg-indent :gps (length goal-stack) "Goal: ~a" goal)
  (cond ((member-equal goal state) state)
        ((member-equal goal goal-stack) nil)
        (t (some #'(lambda (op) (apply-op state goal op goal-stack))
                 (find-all goal *ops* :test #'appropriate-p)))))

(defun member-equal (item list)
  (member item list :test #'equal))

(defun apply-op (state goal op goal-stack)
  "Return a new, transformed state if op is applicable."
  (dbg-indent :gps (length goal-stack) "Consider: ~a" (op-action op))
  (let ((state2 (achieve-all state (op-preconds op) 
                             (cons goal goal-stack))))
    (unless (null state2)
      ;; Return an updated state
      (dbg-indent :gps (length goal-stack) "Action: ~a" (op-action op))
      (append (remove-if #'(lambda (x) 
                             (member-equal x (op-del-list op)))
                         state2)
              (op-add-list op)))))

